@@ -31,15 +31,18 @@
 // ---
 // Author: Sanjay Ghemawat <opensource@google.com>
 
-#include <config.h>
-#ifdef HAVE_INTTYPES_H
+#include "config.h"
+
 #include <inttypes.h>                   // for PRIuPTR
-#endif
 #include <errno.h>                      // for ENOMEM, errno
-#include <gperftools/malloc_extension.h>      // for MallocRange, etc
+
+#include <limits>
+
 #include "base/basictypes.h"
 #include "base/commandlineflags.h"
+#include "gperftools/malloc_extension.h"      // for MallocRange, etc
 #include "internal_logging.h"  // for ASSERT, TCMalloc_Printer, etc
+#include "malloc_backtrace.h"
 #include "page_heap_allocator.h"  // for PageHeapAllocator
 #include "static_vars.h"       // for Static
 #include "system-alloc.h"      // for TCMalloc_SystemAlloc, etc
@@ -62,15 +65,30 @@ DEFINE_int64(tcmalloc_heap_limit_mb,
 
 namespace tcmalloc {
 
-PageHeap::PageHeap()
-    : pagemap_(MetaDataAlloc),
+struct SCOPED_LOCKABLE PageHeap::LockingContext {
+  PageHeap * const heap;
+  size_t grown_by = 0;
+
+  explicit LockingContext(PageHeap* heap, SpinLock* lock) EXCLUSIVE_LOCK_FUNCTION(lock)
+      : heap(heap) {
+    lock->Lock();
+  }
+  ~LockingContext() UNLOCK_FUNCTION() {
+    heap->HandleUnlock(this);
+  }
+};
+
+
+PageHeap::PageHeap(Length smallest_span_size)
+    : smallest_span_size_(smallest_span_size),
+      pagemap_(MetaDataAlloc),
       scavenge_counter_(0),
       // Start scavenging at kMaxPages list
       release_index_(kMaxPages),
       aggressive_decommit_(false) {
-  COMPILE_ASSERT(kClassSizesMax <= (1 << PageMapCache::kValuebits), valuebits);
-  DLL_Init(&large_.normal);
-  DLL_Init(&large_.returned);
+  static_assert(kClassSizesMax <= (1 << PageMapCache::kValuebits));
+  // smallest_span_size needs to be power of 2.
+  CHECK_CONDITION((smallest_span_size_ & (smallest_span_size_-1)) == 0);
   for (int i = 0; i < kMaxPages; i++) {
     DLL_Init(&free_[i].normal);
     DLL_Init(&free_[i].returned);
@@ -78,19 +96,20 @@ PageHeap::PageHeap()
 }
 
 Span* PageHeap::SearchFreeAndLargeLists(Length n) {
+  ASSERT(lock_.IsHeld());
   ASSERT(Check());
   ASSERT(n > 0);
 
   // Find first size >= n that has a non-empty list
-  for (Length s = n; s < kMaxPages; s++) {
-    Span* ll = &free_[s].normal;
+  for (Length s = n; s <= kMaxPages; s++) {
+    Span* ll = &free_[s - 1].normal;
     // If we're lucky, ll is non-empty, meaning it has a suitable span.
     if (!DLL_IsEmpty(ll)) {
       ASSERT(ll->next->location == Span::ON_NORMAL_FREELIST);
       return Carve(ll->next, n);
     }
     // Alternatively, maybe there's a usable returned span.
-    ll = &free_[s].returned;
+    ll = &free_[s - 1].returned;
     if (!DLL_IsEmpty(ll)) {
       // We did not call EnsureLimit before, to avoid releasing the span
       // that will be taken immediately back.
@@ -107,17 +126,58 @@ Span* PageHeap::SearchFreeAndLargeLists(Length n) {
     }
   }
   // No luck in free lists, our last chance is in a larger class.
-  return AllocLarge(n);  // May be NULL
+  return AllocLarge(n);  // May be nullptr
 }
 
 static const size_t kForcedCoalesceInterval = 128*1024*1024;
 
-Span* PageHeap::New(Length n) {
+Length PageHeap::RoundUpSize(Length n) {
+  Length rounded_n = (n + smallest_span_size_ - 1) & ~(smallest_span_size_ - 1);
+  if (rounded_n < n) {
+    // Overflow happened. So make sure we oom by asking for biggest
+    // amount possible.
+    return std::numeric_limits<Length>::max() & ~(smallest_span_size_ - 1);
+  }
+
+  return rounded_n;
+}
+
+void PageHeap::HandleUnlock(LockingContext* context) {
+  StackTrace* t = nullptr;
+  if (context->grown_by) {
+    t = Static::stacktrace_allocator()->New();
+    t->size = context->grown_by;
+  }
+
+  lock_.Unlock();
+
+  if (t) {
+    t->depth = tcmalloc::GrabBacktrace(t->stack, kMaxStackDepth-1, 0);
+    Static::push_growth_stack(t);
+  }
+}
+
+Span* PageHeap::NewWithSizeClass(Length n, uint32_t sizeclass) {
+  LockingContext context{this, &lock_};
+
+  Span* span = NewLocked(n, &context);
+  if (!span) {
+    return span;
+  }
+  InvalidateCachedSizeClass(span->start);
+  if (sizeclass) {
+    RegisterSizeClass(span, sizeclass);
+  }
+  return span;
+}
+
+Span* PageHeap::NewLocked(Length n, LockingContext* context) {
+  ASSERT(lock_.IsHeld());
   ASSERT(Check());
-  ASSERT(n > 0);
+  n = RoundUpSize(n);
 
   Span* result = SearchFreeAndLargeLists(n);
-  if (result != NULL)
+  if (result != nullptr)
     return result;
 
   if (stats_.free_bytes != 0 && stats_.unmapped_bytes != 0
@@ -139,7 +199,7 @@ Span* PageHeap::New(Length n) {
     // penalized with higher count of minor page faults.
     //
     // See also large_heap_fragmentation_unittest.cc and
-    // https://code.google.com/p/gperftools/issues/detail?id=368
+    // https://github.com/gperftools/gperftools/issues/371
     ReleaseAtLeastNPages(static_cast<Length>(0x7fffffff));
 
     // then try again. If we are forced to grow heap because of large
@@ -149,11 +209,11 @@ Span* PageHeap::New(Length n) {
     // unlucky memory fragmentation we'll be consuming virtual address
     // space, but not real memory
     result = SearchFreeAndLargeLists(n);
-    if (result != NULL) return result;
+    if (result != nullptr) return result;
   }
 
   // Grow the heap and try again.
-  if (!GrowHeap(n)) {
+  if (!GrowHeap(n, context)) {
     ASSERT(stats_.unmapped_bytes+ stats_.committed_bytes==stats_.system_bytes);
     ASSERT(Check());
     // underlying SysAllocator likely set ENOMEM but we can get here
@@ -162,51 +222,84 @@ Span* PageHeap::New(Length n) {
     // Setting errno to ENOMEM here allows us to avoid dealing with it
     // in fast-path.
     errno = ENOMEM;
-    return NULL;
+    return nullptr;
   }
   return SearchFreeAndLargeLists(n);
 }
 
+Span* PageHeap::NewAligned(Length n, Length align_pages) {
+  n = RoundUpSize(n);
+
+  // Allocate extra pages and carve off an aligned portion
+  const Length alloc = n + align_pages;
+  if (alloc < n || alloc < align_pages) {
+    // overflow means we asked huge amounts, so lets trigger normal
+    // oom handling by asking enough to trigger oom.
+    Span* span = New(std::numeric_limits<Length>::max());
+    CHECK_CONDITION(span == nullptr);
+    return nullptr;
+  }
+
+  LockingContext context{this, &lock_};
+
+  Span* span = NewLocked(alloc, &context);
+  if (PREDICT_FALSE(span == nullptr)) return nullptr;
+
+  // Skip starting portion so that we end up aligned
+  Length skip = 0;
+  size_t align_bytes = align_pages << kPageShift;
+  while ((((span->start+skip) << kPageShift) & (align_bytes - 1)) != 0) {
+    skip++;
+  }
+  ASSERT(skip < alloc);
+  if (skip > 0) {
+    Span* rest = Split(span, skip);
+    DeleteLocked(span);
+    span = rest;
+  }
+
+  ASSERT(span->length >= n);
+  if (span->length > n) {
+    Span* trailer = Split(span, n);
+    DeleteLocked(trailer);
+  }
+  InvalidateCachedSizeClass(span->start);
+  return span;
+}
+
 Span* PageHeap::AllocLarge(Length n) {
-  // find the best span (closest to n in size).
-  // The following loops implements address-ordered best-fit.
-  Span *best = NULL;
+  ASSERT(lock_.IsHeld());
+  Span *best = nullptr;
+  Span *best_normal = nullptr;
 
-  // Search through normal list
-  for (Span* span = large_.normal.next;
-       span != &large_.normal;
-       span = span->next) {
-    if (span->length >= n) {
-      if ((best == NULL)
-          || (span->length < best->length)
-          || ((span->length == best->length) && (span->start < best->start))) {
-        best = span;
-        ASSERT(best->location == Span::ON_NORMAL_FREELIST);
-      }
-    }
+  // Create a Span to use as an upper bound.
+  Span bound;
+  bound.start = 0;
+  bound.length = n;
+
+  // First search the NORMAL spans..
+  SpanSetIter place = large_normal_.upper_bound(SpanPtrWithLength(&bound));
+  if (place != large_normal_.end()) {
+    best = place->span;
+    best_normal = best;
+    ASSERT(best->location == Span::ON_NORMAL_FREELIST);
   }
 
-  Span *bestNormal = best;
-
-  // Search through released list in case it has a better fit
-  for (Span* span = large_.returned.next;
-       span != &large_.returned;
-       span = span->next) {
-    if (span->length >= n) {
-      if ((best == NULL)
-          || (span->length < best->length)
-          || ((span->length == best->length) && (span->start < best->start))) {
-        best = span;
-        ASSERT(best->location == Span::ON_RETURNED_FREELIST);
-      }
-    }
+  // Try to find better fit from RETURNED spans.
+  place = large_returned_.upper_bound(SpanPtrWithLength(&bound));
+  if (place != large_returned_.end()) {
+    Span *c = place->span;
+    ASSERT(c->location == Span::ON_RETURNED_FREELIST);
+    if (best_normal == nullptr
+        || c->length < best->length)
+      best = place->span;
   }
 
-  if (best == bestNormal) {
-    return best == NULL ? NULL : Carve(best, n);
+  if (best == best_normal) {
+    return best == nullptr ? nullptr : Carve(best, n);
   }
 
-  // best comes from returned list.
+  // best comes from RETURNED set.
 
   if (EnsureLimit(n, false)) {
     return Carve(best, n);
@@ -214,28 +307,27 @@ Span* PageHeap::AllocLarge(Length n) {
 
   if (EnsureLimit(n, true)) {
     // best could have been destroyed by coalescing.
-    // bestNormal is not a best-fit, and it could be destroyed as well.
+    // best_normal is not a best-fit, and it could be destroyed as well.
     // We retry, the limit is already ensured:
     return AllocLarge(n);
   }
 
-  // If bestNormal existed, EnsureLimit would succeeded:
-  ASSERT(bestNormal == NULL);
+  // If best_normal existed, EnsureLimit would succeeded:
+  ASSERT(best_normal == nullptr);
   // We are not allowed to take best from returned list.
-  return NULL;
+  return nullptr;
 }
 
 Span* PageHeap::Split(Span* span, Length n) {
+  ASSERT(lock_.IsHeld());
   ASSERT(0 < n);
   ASSERT(n < span->length);
   ASSERT(span->location == Span::IN_USE);
   ASSERT(span->sizeclass == 0);
-  Event(span, 'T', n);
 
   const int extra = span->length - n;
   Span* leftover = NewSpan(span->start + n, extra);
   ASSERT(leftover->location == Span::IN_USE);
-  Event(leftover, 'U', extra);
   RecordSpan(leftover);
   pagemap_.set(span->start + n - 1, span); // Update map from pageid to span
   span->length = n;
@@ -271,24 +363,22 @@ Span* PageHeap::Carve(Span* span, Length n) {
   const int old_location = span->location;
   RemoveFromFreeList(span);
   span->location = Span::IN_USE;
-  Event(span, 'A', n);
 
   const int extra = span->length - n;
   ASSERT(extra >= 0);
   if (extra > 0) {
     Span* leftover = NewSpan(span->start + n, extra);
     leftover->location = old_location;
-    Event(leftover, 'S', extra);
     RecordSpan(leftover);
 
     // The previous span of |leftover| was just splitted -- no need to
     // coalesce them. The next span of |leftover| was not previously coalesced
-    // with |span|, i.e. is NULL or has got location other than |old_location|.
+    // with |span|, i.e. is nullptr or has got location other than |old_location|.
 #ifndef NDEBUG
     const PageID p = leftover->start;
     const Length len = leftover->length;
     Span* next = GetDescriptor(p+len);
-    ASSERT (next == NULL ||
+    ASSERT (next == nullptr ||
             next->location == Span::IN_USE ||
             next->location != leftover->location);
 #endif
@@ -309,6 +399,12 @@ Span* PageHeap::Carve(Span* span, Length n) {
 }
 
 void PageHeap::Delete(Span* span) {
+  SpinLockHolder h(&lock_);
+  DeleteLocked(span);
+}
+
+void PageHeap::DeleteLocked(Span* span) {
+  ASSERT(lock_.IsHeld());
   ASSERT(Check());
   ASSERT(span->location == Span::IN_USE);
   ASSERT(span->length > 0);
@@ -318,7 +414,6 @@ void PageHeap::Delete(Span* span) {
   span->sizeclass = 0;
   span->sample = 0;
   span->location = Span::ON_NORMAL_FREELIST;
-  Event(span, 'D', span->length);
   MergeIntoFreeList(span);  // Coalesces if possible
   IncrementalScavenge(n);
   ASSERT(stats_.unmapped_bytes+ stats_.committed_bytes==stats_.system_bytes);
@@ -329,9 +424,9 @@ void PageHeap::Delete(Span* span) {
 // checks if 'other' span is mergable with 'span'. If it is, removes
 // other span from free list, performs aggressive decommit if
 // necessary and returns 'other' span. Otherwise 'other' span cannot
-// be merged and is left untouched. In that case NULL is returned.
+// be merged and is left untouched. In that case nullptr is returned.
 Span* PageHeap::CheckAndHandlePreMerge(Span* span, Span* other) {
-  if (other == NULL) {
+  if (other == nullptr) {
     return other;
   }
   // if we're in aggressive decommit mode and span is decommitted,
@@ -340,10 +435,10 @@ Span* PageHeap::CheckAndHandlePreMerge(Span* span, Span* other) {
       && span->location == Span::ON_RETURNED_FREELIST) {
     bool worked = DecommitSpan(other);
     if (!worked) {
-      return NULL;
+      return nullptr;
     }
   } else if (other->location != span->location) {
-    return NULL;
+    return nullptr;
   }
 
   RemoveFromFreeList(other);
@@ -351,6 +446,7 @@ Span* PageHeap::CheckAndHandlePreMerge(Span* span, Span* other) {
 }
 
 void PageHeap::MergeIntoFreeList(Span* span) {
+  ASSERT(lock_.IsHeld());
   ASSERT(span->location != Span::IN_USE);
 
   // Coalesce -- we guarantee that "p" != 0, so no bounds checking
@@ -380,7 +476,7 @@ void PageHeap::MergeIntoFreeList(Span* span) {
   }
 
   Span* prev = CheckAndHandlePreMerge(span, GetDescriptor(p-1));
-  if (prev != NULL) {
+  if (prev != nullptr) {
     // Merge preceding span into this span
     ASSERT(prev->start + prev->length == p);
     const Length len = prev->length;
@@ -388,45 +484,70 @@ void PageHeap::MergeIntoFreeList(Span* span) {
     span->start -= len;
     span->length += len;
     pagemap_.set(span->start, span);
-    Event(span, 'L', len);
   }
   Span* next = CheckAndHandlePreMerge(span, GetDescriptor(p+n));
-  if (next != NULL) {
+  if (next != nullptr) {
     // Merge next span into this span
     ASSERT(next->start == p+n);
     const Length len = next->length;
     DeleteSpan(next);
     span->length += len;
     pagemap_.set(span->start + span->length - 1, span);
-    Event(span, 'R', len);
   }
 
   PrependToFreeList(span);
 }
 
 void PageHeap::PrependToFreeList(Span* span) {
+  ASSERT(lock_.IsHeld());
   ASSERT(span->location != Span::IN_USE);
-  SpanList* list = (span->length < kMaxPages) ? &free_[span->length] : &large_;
-  if (span->location == Span::ON_NORMAL_FREELIST) {
+  if (span->location == Span::ON_NORMAL_FREELIST)
     stats_.free_bytes += (span->length << kPageShift);
+  else
+    stats_.unmapped_bytes += (span->length << kPageShift);
+
+  if (span->length > kMaxPages) {
+    SpanSet *set = &large_normal_;
+    if (span->location == Span::ON_RETURNED_FREELIST)
+      set = &large_returned_;
+    std::pair<SpanSetIter, bool> p =
+        set->insert(SpanPtrWithLength(span));
+    ASSERT(p.second); // We never have duplicates since span->start is unique.
+    span->SetSpanSetIterator(p.first);
+    return;
+  }
+
+  SpanList* list = &free_[span->length - 1];
+  if (span->location == Span::ON_NORMAL_FREELIST) {
     DLL_Prepend(&list->normal, span);
   } else {
-    stats_.unmapped_bytes += (span->length << kPageShift);
     DLL_Prepend(&list->returned, span);
   }
 }
 
 void PageHeap::RemoveFromFreeList(Span* span) {
+  ASSERT(lock_.IsHeld());
   ASSERT(span->location != Span::IN_USE);
   if (span->location == Span::ON_NORMAL_FREELIST) {
     stats_.free_bytes -= (span->length << kPageShift);
   } else {
     stats_.unmapped_bytes -= (span->length << kPageShift);
   }
-  DLL_Remove(span);
+  if (span->length > kMaxPages) {
+    SpanSet *set = &large_normal_;
+    if (span->location == Span::ON_RETURNED_FREELIST)
+      set = &large_returned_;
+    SpanSetIter iter = span->ExtractSpanSetIterator();
+    ASSERT(iter->span == span);
+    ASSERT(set->find(SpanPtrWithLength(span)) == iter);
+    set->erase(iter);
+  } else {
+    DLL_Remove(span);
+  }
 }
 
 void PageHeap::IncrementalScavenge(Length n) {
+  ASSERT(lock_.IsHeld());
   // Fast path; not yet time to release memory
   scavenge_counter_ -= n;
   if (scavenge_counter_ >= 0) return;  // Not yet time to scavenge
@@ -459,8 +580,7 @@ void PageHeap::IncrementalScavenge(Length n) {
   }
 }
 
-Length PageHeap::ReleaseLastNormalSpan(SpanList* slist) {
-  Span* s = slist->normal.prev;
+Length PageHeap::ReleaseSpan(Span* s) {
   ASSERT(s->location == Span::ON_NORMAL_FREELIST);
 
   if (DecommitSpan(s)) {
@@ -475,30 +595,45 @@ Length PageHeap::ReleaseLastNormalSpan(SpanList* slist) {
 }
 
 Length PageHeap::ReleaseAtLeastNPages(Length num_pages) {
+  ASSERT(lock_.IsHeld());
   Length released_pages = 0;
 
-  // Round robin through the lists of free spans, releasing the last
-  // span in each list.  Stop after releasing at least num_pages
+  // Round robin through the lists of free spans, releasing a
+  // span from each list.  Stop after releasing at least num_pages
   // or when there is nothing more to release.
   while (released_pages < num_pages && stats_.free_bytes > 0) {
     for (int i = 0; i < kMaxPages+1 && released_pages < num_pages;
          i++, release_index_++) {
+      Span *s;
       if (release_index_ > kMaxPages) release_index_ = 0;
-      SpanList* slist = (release_index_ == kMaxPages) ?
-          &large_ : &free_[release_index_];
-      if (!DLL_IsEmpty(&slist->normal)) {
-        Length released_len = ReleaseLastNormalSpan(slist);
-        // Some systems do not support release
-        if (released_len == 0) return released_pages;
-        released_pages += released_len;
+
+      if (release_index_ == kMaxPages) {
+        if (large_normal_.empty()) {
+          continue;
+        }
+        s = (large_normal_.begin())->span;
+      } else {
+        SpanList* slist = &free_[release_index_];
+        if (DLL_IsEmpty(&slist->normal)) {
+          continue;
+        }
+        s = slist->normal.prev;
       }
+      // TODO(todd) if the remaining number of pages to release
+      // is significantly smaller than s->length, and s is on the
+      // large freelist, should we carve s instead of releasing?
+      // the whole thing?
+      Length released_len = ReleaseSpan(s);
+      // Some systems do not support release
+      if (released_len == 0) return released_pages;
+      released_pages += released_len;
     }
   }
   return released_pages;
 }
 
-bool PageHeap::EnsureLimit(Length n, bool withRelease)
-{
+bool PageHeap::EnsureLimit(Length n, bool withRelease) {
+  ASSERT(lock_.IsHeld());
   Length limit = (FLAGS_tcmalloc_heap_limit_mb*1024*1024) >> kPageShift;
   if (limit == 0) return true; //there is no limit
 
@@ -521,42 +656,44 @@ bool PageHeap::EnsureLimit(Length n, bool withRelease)
   return takenPages + n <= limit;
 }
 
-void PageHeap::RegisterSizeClass(Span* span, uint32 sc) {
+void PageHeap::RegisterSizeClass(Span* span, uint32_t sc) {
   // Associate span object with all interior pages as well
   ASSERT(span->location == Span::IN_USE);
   ASSERT(GetDescriptor(span->start) == span);
   ASSERT(GetDescriptor(span->start+span->length-1) == span);
-  Event(span, 'C', sc);
   span->sizeclass = sc;
   for (Length i = 1; i < span->length-1; i++) {
     pagemap_.set(span->start+i, span);
   }
 }
 
-void PageHeap::GetSmallSpanStats(SmallSpanStats* result) {
-  for (int s = 0; s < kMaxPages; s++) {
-    result->normal_length[s] = DLL_Length(&free_[s].normal);
-    result->returned_length[s] = DLL_Length(&free_[s].returned);
+void PageHeap::GetSmallSpanStatsLocked(SmallSpanStats* result) {
+  ASSERT(lock_.IsHeld());
+  for (int i = 0; i < kMaxPages; i++) {
+    result->normal_length[i] = DLL_Length(&free_[i].normal);
+    result->returned_length[i] = DLL_Length(&free_[i].returned);
   }
 }
 
-void PageHeap::GetLargeSpanStats(LargeSpanStats* result) {
+void PageHeap::GetLargeSpanStatsLocked(LargeSpanStats* result) {
+  ASSERT(lock_.IsHeld());
   result->spans = 0;
   result->normal_pages = 0;
   result->returned_pages = 0;
-  for (Span* s = large_.normal.next; s != &large_.normal; s = s->next) {
-    result->normal_pages += s->length;;
+  for (SpanSetIter it = large_normal_.begin(); it != large_normal_.end(); ++it) {
+    result->normal_pages += it->length;
     result->spans++;
   }
-  for (Span* s = large_.returned.next; s != &large_.returned; s = s->next) {
-    result->returned_pages += s->length;
+  for (SpanSetIter it = large_returned_.begin(); it != large_returned_.end(); ++it) {
+    result->returned_pages += it->length;
     result->spans++;
   }
 }
 
 bool PageHeap::GetNextRange(PageID start, base::MallocRange* r) {
+  ASSERT(lock_.IsHeld());
   Span* span = reinterpret_cast<Span*>(pagemap_.Next(start));
-  if (span == NULL) {
+  if (span == nullptr) {
     return false;
   }
   r->address = span->start << kPageShift;
@@ -585,24 +722,17 @@ bool PageHeap::GetNextRange(PageID start, base::MallocRange* r) {
   return true;
 }
 
-static void RecordGrowth(size_t growth) {
-  StackTrace* t = Static::stacktrace_allocator()->New();
-  t->depth = GetStackTrace(t->stack, kMaxStackDepth-1, 3);
-  t->size = growth;
-  t->stack[kMaxStackDepth-1] = reinterpret_cast<void*>(Static::growth_stacks());
-  Static::set_growth_stacks(t);
-}
-
-bool PageHeap::GrowHeap(Length n) {
+bool PageHeap::GrowHeap(Length n, LockingContext* context) {
+  ASSERT(lock_.IsHeld());
   ASSERT(kMaxPages >= kMinSystemAlloc);
   if (n > kMaxValidPages) return false;
   Length ask = (n>kMinSystemAlloc) ? n : static_cast<Length>(kMinSystemAlloc);
   size_t actual_size;
-  void* ptr = NULL;
+  void* ptr = nullptr;
   if (EnsureLimit(ask)) {
       ptr = TCMalloc_SystemAlloc(ask << kPageShift, &actual_size, kPageSize);
   }
-  if (ptr == NULL) {
+  if (ptr == nullptr) {
     if (n < ask) {
       // Try growing just "n" pages
       ask = n;
@@ -610,10 +740,10 @@ bool PageHeap::GrowHeap(Length n) {
         ptr = TCMalloc_SystemAlloc(ask << kPageShift, &actual_size, kPageSize);
       }
     }
-    if (ptr == NULL) return false;
+    if (ptr == nullptr) return false;
   }
   ask = actual_size >> kPageShift;
-  RecordGrowth(ask << kPageShift);
+  context->grown_by += ask << kPageShift;
 
   ++stats_.reserve_count;
   ++stats_.commit_count;
@@ -645,7 +775,7 @@ bool PageHeap::GrowHeap(Length n) {
     // any necessary coalescing to occur.
     Span* span = NewSpan(p, ask);
     RecordSpan(span);
-    Delete(span);
+    DeleteLocked(span);
     ASSERT(stats_.unmapped_bytes+ stats_.committed_bytes==stats_.system_bytes);
     ASSERT(Check());
     return true;
@@ -657,18 +787,17 @@ bool PageHeap::GrowHeap(Length n) {
 }
 
 bool PageHeap::Check() {
-  ASSERT(free_[0].normal.next == &free_[0].normal);
-  ASSERT(free_[0].returned.next == &free_[0].returned);
+  ASSERT(lock_.IsHeld());
   return true;
 }
 
 bool PageHeap::CheckExpensive() {
   bool result = Check();
-  CheckList(&large_.normal, kMaxPages, 1000000000, Span::ON_NORMAL_FREELIST);
-  CheckList(&large_.returned, kMaxPages, 1000000000, Span::ON_RETURNED_FREELIST);
-  for (Length s = 1; s < kMaxPages; s++) {
-    CheckList(&free_[s].normal, s, s, Span::ON_NORMAL_FREELIST);
-    CheckList(&free_[s].returned, s, s, Span::ON_RETURNED_FREELIST);
+  CheckSet(&large_normal_, kMaxPages + 1, Span::ON_NORMAL_FREELIST);
+  CheckSet(&large_returned_, kMaxPages + 1, Span::ON_RETURNED_FREELIST);
+  for (int s = 1; s <= kMaxPages; s++) {
+    CheckList(&free_[s - 1].normal, s, s, Span::ON_NORMAL_FREELIST);
+    CheckList(&free_[s - 1].returned, s, s, Span::ON_RETURNED_FREELIST);
   }
   return result;
 }
@@ -685,4 +814,16 @@ bool PageHeap::CheckList(Span* list, Length min_pages, Length max_pages,
   return true;
 }
 
+bool PageHeap::CheckSet(SpanSet* spanset, Length min_pages,int freelist) {
+  for (SpanSetIter it = spanset->begin(); it != spanset->end(); ++it) {
+    Span* s = it->span;
+    CHECK_CONDITION(s->length == it->length);
+    CHECK_CONDITION(s->location == freelist);  // NORMAL or RETURNED
+    CHECK_CONDITION(s->length >= min_pages);
+    CHECK_CONDITION(GetDescriptor(s->start) == s);
+    CHECK_CONDITION(GetDescriptor(s->start+s->length-1) == s);
+  }
+  return true;
+}
+
 }  // namespace tcmalloc
